/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CMEMORY_CONTAINER_HPP
#define CMEMORY_CONTAINER_HPP

#include <iostream>

/***************
 * PUSH
 ***************/
#define CONTAINER_PUSH(var_type)						\
	void push(var_type value)						\
	{									\
		if (dataSize + sizeof(var_type) >= containerSize) 		\
		{								\
			containerSize += blockSize;				\
			containerData = (char*)realloc(containerData, containerSize);	\
			dataPtr = containerData;				\
		}								\
		*(var_type*)(containerData+dataSize) = value;			\
		dataSize += sizeof(var_type[1]);				\
	}									\
										\
	void push(int n, var_type *value)					\
	{									\
		if (dataSize + sizeof(var_type) >= containerSize) 		\
		{								\
			containerSize += blockSize;				\
			containerData = (char*)realloc(containerData, containerSize);	\
			dataPtr = containerData;				\
		}								\
		for(int i = 0; i < n; i++)					\
			*(var_type*)(containerData+dataSize+sizeof(var_type)*i) = value[i];		\
		dataSize += sizeof(var_type)*n;				\
	}

#define CONTAINER_PUSH_ARRAY_2(var_type)					\
	void push2(var_type value[2])						\
	{									\
		if (dataSize + sizeof(var_type[2]) >= containerSize) 		\
		{								\
			containerSize += blockSize;				\
			containerData = (char*)realloc(containerData, containerSize);	\
			dataPtr = containerData;				\
		}								\
		*(var_type*)(containerData+dataSize) = value[0];		\
		*(var_type*)(containerData+dataSize+sizeof(var_type[1])) = value[1];		\
		dataSize += sizeof(var_type[2]);				\
	}

#define CONTAINER_PUSH_ARRAY_3(var_type)					\
	void push3(var_type value[3])						\
	{									\
		if (dataSize + sizeof(var_type[3]) >= containerSize) 		\
		{								\
			containerSize += blockSize;				\
			containerData = (char*)realloc(containerData, containerSize);	\
			dataPtr = containerData;				\
		}								\
		*(var_type*)(containerData+dataSize) = value[0];		\
		*(var_type*)(containerData+dataSize+sizeof(var_type[1])) = value[1];		\
		*(var_type*)(containerData+dataSize+sizeof(var_type[2])) = value[2];		\
		dataSize += sizeof(var_type[3]);				\
	}


/***************
 * POP data and decrease container size
 ***************/
#define CONTAINER_POP(var_type)							\
	void pop(var_type *ret)							\
	{									\
		unsigned int newDataSize = dataSize - sizeof(var_type);		\
		*ret = *(var_type*)(containerData+newDataSize);			\
		if (newDataSize <= containerSize - blockSize)			\
		{								\
			containerSize -= blockSize;				\
			containerData = (char*)realloc(containerData, containerSize);	\
			dataPtr = containerData;				\
		}								\
		dataSize = newDataSize;						\
	}

#define CONTAINER_POP_ARRAY_2(var_type)						\
	void pop2(var_type *ret)						\
	{									\
		unsigned int newDataSize = dataSize - sizeof(var_type[2]);	\
		ret[0] = *(var_type*)(containerData+newDataSize);		\
		ret[1] = *(var_type*)(containerData+newDataSize+sizeof(var_type[1]));		\
		if (newDataSize <= containerSize - blockSize)			\
		{								\
			containerSize -= blockSize;				\
			containerData = (char*)realloc(containerData, containerSize);	\
			dataPtr = containerData;				\
		}								\
		dataSize = newDataSize;						\
	}

#define CONTAINER_POP_ARRAY_3(var_type)						\
	void pop3(var_type *ret)						\
	{									\
		unsigned int newDataSize = dataSize - sizeof(var_type[3]);	\
		ret[0] = *(var_type*)(containerData+newDataSize);		\
		ret[1] = *(var_type*)(containerData+newDataSize+sizeof(var_type[1]));	\
		ret[2] = *(var_type*)(containerData+newDataSize+sizeof(var_type[2]));	\
		if (newDataSize <= containerSize - blockSize)			\
		{								\
			containerSize -= blockSize;				\
			containerData = (char*)realloc(containerData, containerSize);	\
			dataPtr = containerData;				\
		}								\
		dataSize = newDataSize;						\
	}


/***************
 * GET data without decreasing the container size
 ***************/
#define CONTAINER_GET(var_type)							\
	void get(var_type *ret)							\
	{									\
		*ret = *(var_type*)dataPtr;					\
		dataPtr += sizeof(var_type[1]);					\
	}									\
										\
	void get(int n, var_type *ret)						\
	{									\
		for (int i = 0; i < n; i++)					\
			ret[i] = *(var_type*)(dataPtr+sizeof(var_type)*i);	\
		dataPtr += sizeof(var_type)*n;					\
	}

#define CONTAINER_GET_ARRAY_2(var_type)						\
	void get2(var_type *ret)						\
	{									\
		ret[0] = *(var_type*)(dataPtr);		\
		ret[1] = *(var_type*)(dataPtr+sizeof(var_type[1]));		\
		dataPtr += sizeof(var_type[2]);					\
	}

#define CONTAINER_GET_ARRAY_3(var_type)						\
	void get3(var_type *ret)						\
	{									\
		ret[0] = *(var_type*)(dataPtr);		\
		ret[1] = *(var_type*)(dataPtr+sizeof(var_type[1]));		\
		ret[2] = *(var_type*)(dataPtr+sizeof(var_type[2]));		\
		dataPtr += sizeof(var_type[3]);					\
	}





class CMemoryContainer
{
public:
	char *containerData;
	unsigned int dataSize;
	unsigned int containerSize;
	unsigned int blockSize;
	char *dataPtr;

public:

	CONTAINER_PUSH(int)
	CONTAINER_PUSH(float)
	CONTAINER_PUSH(double)
	CONTAINER_PUSH_ARRAY_2(int)
	CONTAINER_PUSH_ARRAY_2(float)
	CONTAINER_PUSH_ARRAY_2(double)
	CONTAINER_PUSH_ARRAY_3(int)
	CONTAINER_PUSH_ARRAY_3(float)
	CONTAINER_PUSH_ARRAY_3(double)

	CONTAINER_POP(int)
	CONTAINER_POP(float)
	CONTAINER_POP(double)
	CONTAINER_POP_ARRAY_2(int)
	CONTAINER_POP_ARRAY_2(float)
	CONTAINER_POP_ARRAY_2(double)
	CONTAINER_POP_ARRAY_3(int)
	CONTAINER_POP_ARRAY_3(float)
	CONTAINER_POP_ARRAY_3(double)

	CONTAINER_GET(int)
	CONTAINER_GET(float)
	CONTAINER_GET(double)
	CONTAINER_GET_ARRAY_2(int)
	CONTAINER_GET_ARRAY_2(float)
	CONTAINER_GET_ARRAY_2(double)
	CONTAINER_GET_ARRAY_3(int)
	CONTAINER_GET_ARRAY_3(float)
	CONTAINER_GET_ARRAY_3(double)

	CMemoryContainer(int p_blockSize = 1024*1024)
	{
		dataSize = 0;
		containerSize = 0;
		blockSize = 0;
		blockSize = p_blockSize;
		containerData = NULL;
		dataPtr = NULL;
		if (blockSize < sizeof(float)*4)
			std::cerr << "warning: blockSize < " << (int)(sizeof(float)*4) << std::endl;
	}

	unsigned int size()
	{
		return dataSize;
	}

	void *data()
	{
		return (void*)containerData;
	}

	void setSize(unsigned int size)
	{
		containerData = (char*)realloc(containerData, size);
		dataPtr = containerData;
		dataSize = size;
		containerSize = size;
	}

	~CMemoryContainer()
	{
		if (containerData)
			free(containerData);
	}

	void clear()
	{
		free(containerData);
		containerSize = 0;
		dataSize = 0;
		dataPtr = NULL;
		containerData = NULL;
	}
};

#endif
